local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")

function Capitalist:ApplyFilters()
	self:DebugEntry("ApplyFilters")

	-- reset memory
	self:ComponentMemoryReset()
	self:ProductMemoryReset()
	self:CostMemoryReset()

	-- REVIEW: Enable seeing and setting counts for crafts by alts
	-- CONSIDER:  Automatic queueing of subcomponents that alts need to make

	local productsToProcess = {}
	local componentsToMakeVisible = {}

	local q = 0

	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			productsToProcess[profession] = { skills = {}, pending = 1, }

			self:DebugMisc(profession)

			for skillName, skillData in pairs(professionData) do

				-- REVIEW: Does this get called in scenarios such as finding a new product?  If not, add helper function etc.
				--
				-- products are not allowed to have costs set directly
				-- update data in case products were discovered after
				-- costs were set on the product (as a component)
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (not self:SkillIsCyclic(profession, skillName) and
					self:IsComponent(productName) and
					self.db.profile.components[productName].cost) then

					local cost = self:GetComponentCost(productName) * 10000
					local costString = self:PrettyCurrency(cost)
					self:Print(string.format("  "..L["Removing cost (%s) from product (%s)."], costString, self:GetComponentText(productName)))
					self:SetComponentCost(productName, nil)
				end

				-- for products not allowed, mark them as blocked
				local productBlocked
				if (not self:ProductIsAllowed(profession, skillName)) then
					productBlocked = true
				end

				-- for products with unavailable product name, block as well
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (not productName) then
					productBlocked = true
				end

				self:ProductSetVisible(profession, skillName, not productBlocked)
				if (not productBlocked) then
					if (self:ProductIsEnabled(profession, skillName)) then
						table.insert(productsToProcess[profession].skills, skillName)
					end
				end
			end
		end
	end

	local doneWithProducts
	while (not doneWithProducts) do
		doneWithProducts = true
		for profession, professionData in pairs(self.db.profile.professions) do
			if (productsToProcess[profession].skills) then
				while (productsToProcess[profession].pending <= #productsToProcess[profession].skills) do
					local skillName = productsToProcess[profession].skills[productsToProcess[profession].pending]
					local product = self.db.profile.professions[profession][skillName]
					for componentName, componentCount in pairs(product.components) do
						local componentSkillName, componentProfession = self:SkillNameFromProductName(componentName)
						if (componentSkillName and componentProfession) then
							if (not tContains(productsToProcess[componentProfession].skills, componentSkillName)) then
								self:ProductSetVisible(componentProfession, componentSkillName, true)
								table.insert(productsToProcess[componentProfession].skills, componentSkillName)
								doneWithProducts = false
							end
						else
							if (not tContains(componentsToMakeVisible, componentName)) then
								table.insert(componentsToMakeVisible, componentName)
							end
						end
					end

					-- enchanting-specific
					local productName = self:ProductNameFromSkillName(profession, skillName)
					if (productName:find(L["Scroll of Enchant "])) then
						local componentName = L["Enchanting Vellum"]
						local componentSkillName, componentProfession = self:SkillNameFromProductName(componentName)
						if (not tContains(productsToProcess[componentProfession].skills, componentSkillName)) then
							self:ProductSetVisible(componentProfession, componentSkillName, true)
							table.insert(productsToProcess[componentProfession].skills, componentSkillName)
							doneWithProducts = false
						end
					end

					productsToProcess[profession].pending = productsToProcess[profession].pending + 1
				end
			end
		end
	end

	for i=1, #componentsToMakeVisible do
		-- make visible components that are in use
		local componentName = componentsToMakeVisible[i]
		if (not self:ComponentGetMerchant(componentName)) then
			self:ComponentSetVisible(componentName, true)
		end
	end

	self:BuildList()
	self:UpdateOptions()

	self:DebugExit("ApplyFilters")
end

-- scans the trade skill frame and adds all products the player knows
function Capitalist:ScanTradeSkillFrame(...)
	self:DebugEntry("ScanTradeSkillFrame")

	local profession = ...;
	self:DebugMisc(profession)

	-- REVIEW: Update products for other players, and maybe track their names to do business later
	-- Bypass data updating when viewing another player's profession information
	if (IsTradeSkillLinked()) then
	 return
	end

	SetTradeSkillSubClassFilter(0, 1, 1)
	SetTradeSkillInvSlotFilter(0, 1, 1)

	-- REVIEW: This had errors for non-Enchanting professions
	--ExpandTradeSkillSubClass(1)

	SetTradeSkillItemNameFilter("")
	
	local newCharacterProducts = 0

	if (not self.db.profile[realm]) then
		self.db.profile[realm] = {}
	end
	if (not self.db.profile[realm][character]) then
		self.db.profile[realm][character] = {}
	end
	if (not self.db.profile[realm][character][profession]) then
		self.db.profile[realm][character][profession] = {}
	end

	local newProducts = 0
	local newComponents = 0

	if (not self.db.profile.professions) then
		self.db.profile.professions = {}
	end
	if (not self.db.profile.professions[profession]) then
		self.db.profile.professions[profession] = {}
	end
	
	if (not self.db.profile.components) then
		self.db.profile.components = {}
	end
	
	for skillId=1, GetNumTradeSkills() do
		local skillName, skillType, numAvailable, isExpanded, altVerb = GetTradeSkillInfo(skillId)
		self:DebugMisc(skillName)		-- "Enchant Bracers - Major Stamina"
		self:DebugMisc(skillType)		-- "header" or one of ("trivial", "easy", "medium", "optimal", "difficult")
		self:DebugMisc(numAvailable)	-- the bracketed value shown in the craft UI, showing how many you can make with current components
		self:DebugMisc(isExpanded)		--
		self:DebugMisc(altVerb)			-- nil or "Enchant" (REVIEW: perhaps others for armor kits?)

		local itemLink = GetTradeSkillItemLink(skillId)
		self:DebugMisc(itemLink)

		local recipeLink = GetTradeSkillRecipeLink(skillId)
		self:DebugMisc(recipeLink)

		if (skillType ~= "header" and L.checkForValidSkillName(profession, skillName)) then
			local skillDescription = GetTradeSkillDescription(skillId)
			self:DebugMisc(skillDescription)

			local minMade, maxMade = GetTradeSkillNumMade(skillId)
			self:DebugMisc(minMade)
			self:DebugMisc(maxMade)

			if (maxMade == minMade) then
				maxMade = nil
				if (minMade == 1) then
					minMade = nil
				end
			end

			-- add it to the known products
			if (not self.db.profile.professions[profession][skillName]) then
				self.db.profile.professions[profession][skillName] = {
																		components = {},
																	 }

				newProducts = newProducts + 1
			end

			-- add the known products for this character
			if (not self.db.profile[realm][character][profession][skillName]) then
				self.db.profile[realm][character][profession][skillName] = { }
				newCharacterProducts = newCharacterProducts + 1
			end


			-- CONSIDER: For all information extractabl from link, itemLink, or recipeLink *don't store it*.

			if (self.db.profile.professions[profession][skillName].description ~= skillDescription) then
				self.db.profile.professions[profession][skillName].description = skillDescription
			end

			if (self.db.profile.professions[profession][skillName].itemLink ~= itemLink) then
				self.db.profile.professions[profession][skillName].itemLink = itemLink
			end

			if (self.db.profile.professions[profession][skillName].recipeLink ~= recipeLink) then
				self.db.profile.professions[profession][skillName].recipeLink = recipeLink
			end

			if (self.db.profile.professions[profession][skillName].min ~= minMade) then
				self.db.profile.professions[profession][skillName].min = minMade
			end
			if (self.db.profile.professions[profession][skillName].max ~= maxMade) then
				self.db.profile.professions[profession][skillName].max = maxMade
			end

			if (self.db.profile[realm][character][profession][skillName].type ~= skillType) then
				self.db.profile[realm][character][profession][skillName].type = skillType
			end

			-- TODO: Improve components updating to only apply when there are changes [PERF]
			self.db.profile.professions[profession][skillName].components = {}

			for reagentId=1, GetTradeSkillNumReagents(skillId) do
				local reagentName, reagentTexture, reagentCount, playerReagentCount = GetTradeSkillReagentInfo(skillId, reagentId)
				self:DebugMisc(reagentName)
				self:DebugMisc(reagentTexture)
				self:DebugMisc(reagentCount)
				self:DebugMisc(playerReagentCount)

				local reagentLink = GetTradeSkillReagentItemLink(skillId, reagentId)
				self:DebugMisc(reagentLink)

				if (reagentName and (reagentCount > 0)) then
					self.db.profile.professions[profession][skillName].components[reagentName] = reagentCount

					-- for previously unknown components, save the item id
					if (not self.db.profile.components[reagentName]) then
						local reagentLink = GetTradeSkillReagentItemLink(skillId, reagentId)
						self:DebugMisc(reagentLink)

						local _, _, reagentItemId = reagentLink:find("Hitem:(%d+):")
						self:DebugMisc(reagentItemId)
					
						self.db.profile.components[reagentName] = {
																	id = tonumber(reagentItemId)
																  }

						newComponents = newComponents + 1
					end

					if (self.db.profile.components[reagentName].link ~= reagentLink) then
						self.db.profile.components[reagentName].link = reagentLink
					end
				end
			end
		end
	end

	if (profession == L["Enchanting"]) then
		local componentName = L["Enchanting Vellum"]
		if (not self.db.profile.components[componentName]) then
			self.db.profile.components[componentName] = {
															id = self:GetComponentId(componentName),
														}
			newComponents = newComponents + 1
		end
	end
	
	if (newProducts + newComponents + newCharacterProducts > 0) then
		self:Print(string.format(
						L["Found %d new products and %d new components."],
						newProducts,
						newComponents
					))
		self:Print(string.format(
						L["Found %d new character products."],
						newCharacterProducts
					))

		self:ApplyFilters()
		self:UpdateOptions()
	end
	
	self:UpdateRestockStack(profession)

	self:DebugExit("ScanTradeSkillFrame")
end

function Capitalist:BuildList()
	self:DebugEntry("BuildList")

	self.list = {}

	-- Generally, it's "simple bucketing" -- 1 full product name per bucket
	self.scanBuckets = {}
	-- Optimization 1: do "Glyph of " in batch
	local productsIncludeGlyph = false
	-- Optimization 2: do "Scroll of Enchant " in batch
	local productsIncludeScroll = false

	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			self:DebugMisc(profession)

			for skillName, skillData in pairs(professionData) do
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (productName) then
					local applyOptimizationGlyph = false
					local applyOptimizationScroll = false
					if (productName:find(L["Glyph of "])) then
						applyOptimizationGlyph = true
					end
					if (productName:find(L["Scroll of Enchant "])) then
						applyOptimizationScroll = true
					end

					if (productName and
						self:ProductIsVisible(profession, skillName)) then

						self.list[productName] = { profession = profession, skillName = skillName }

						if (self:ProductIsEnabled(profession, skillName)) then
							if (applyOptimizationGlyph) then
								productsIncludeGlyph = true
							elseif (applyOptimizationScroll) then
								productsIncludeScroll = true
							else
								table.insert(self.scanBuckets, productName)
							end
						end
					end
				end
			end
		end
	end

	if (productsIncludeGlyph) then
		table.insert(self.scanBuckets, 1, L["Glyph of "])
	end
	if (productsIncludeScroll) then
		table.insert(self.scanBuckets, 1, L["Scroll of Enchant "])
	end

	self:DebugExit("BuildList")
end
